:- module(explore, [explore/8]).

/* <module> Explore Algorithm
 *
 *  This module provides the predicate explore/8, that
 *  implements the algorithm given in the paper
 *  "Algorithmic Aspects of Theory Blending".
 *  It computes maximal consistent and minimal inconsistent
 *  theories in a lattice formed by combining formulae from two
 *  input theories.
 *
 *  This module expects a predicate consistent/1 to be defined,
 *  that will be used to check theories for consistency.
 *
 *  @title Explore
 *  @author Ulf Krumnack
 *  @version 1.0
 *  @license GPL
 */



%! element_of_set(+Element:term, +Set:list) is semidet
%
% Check if Element is an element of the set Set.
element_of_set(_, []) :- !, fail.
    
element_of_set(E, [E|_]) :- !.

element_of_set(E, [_|R]) :- !,
    element_of_set(E, R).


%! remove_element(+Element:term, +SetIn:list, -SetOut:list) is det
%
% True if SetOut is a SetIn with the first occurence of Element removed.
remove_element(_, [], []) :- !.

remove_element(E, [E|R], R) :- !.

remove_element(E, [_|R], R2) :-
    remove_element(E, R, R2).


%! subset_of_set(+Subset:list, Set:list) is semidet
%
% True if Subset is a subset of Set, i.e. if every element of Subset
% is also an element of Set.
subset_of_set([], _) :- !.

subset_of_set([E|R], Set) :-
    element_of_set(E,Set),
    subset_of_set(R, Set).


%! subset_of_sets(+Subset:list, +Sets:list_of_lists) is semidet
%
% True if Subset is a subset of an element of Sets (which is as
% set of sets), i.e. if there exists a Set in Sets, in which Subset
% is contained.
subset_of_sets(_, []) :- !, fail.

subset_of_sets(Subset, [Set|_]) :-
    subset_of_set(Subset,Set), !.

subset_of_sets(Subset, [_|Sets]) :-
    subset_of_sets(Subset,Sets).


%! superset_of_sets(+Superset:list, +Sets:list_of_lists) is semidet
%
% True if Superset is a superset of an element of Sets
% (which is as set of sets), i.e. if there exists a Set in Sets,
% which is a subset of Superset.
superset_of_sets(_, []) :- !, fail.

superset_of_sets(Superset, [Set|_]) :-
    subset_of_set(Set, Superset), !.

superset_of_sets(Superset, [_|Sets]) :-
    superset_of_sets(Superset,Sets).


%! subset_extend(+SetsIn:list_of_list, +Set:list, -SetsOut:list_of_lists) is det
%
% Extend a collection of non-mutually including sets (SetsIn).
% If the new Set is a subset of a member of the collection, this member
% will be removed.
% It is assumed (and therefore not checked) that the newly inserted set
% is not a superset of any set in the collection.
subset_extend([], Subset, [Subset]) :- !.

subset_extend([Set|Sets], Subset, Sets2) :-
    subset(Subset, Set), !,
    subset_extend(Sets, Subset, Sets2).

subset_extend([Set|Sets], Subset, [Set|Sets2]) :-
    subset_extend(Sets, Subset, Sets2).


%! superset_extend(+SetsIn:list_of_lists, +Superset:list, -SetsOut:list_of_lists) is det
%
% Extend a collection of non-mutually including sets (SetsIn).
% If the new Set is a superset of a member of the collection, this member
% will be removed.
% It is assumed (and therefore not checked) that the newly inserted set
% is not a subset of any set in the collection.
superset_extend([], Superset, [Superset]) :- !.

superset_extend([Set|Sets], Superset, Sets2) :-
    subset(Set,Superset), !,
    superset_extend(Sets, Superset, Sets2).

superset_extend([Set|Sets], Superset, [Set|Sets2]) :-
    superset_extend(Sets, Superset, Sets2).



%! explore(+Base:list, +Axioms:list, +Theory:list, +MaxConIn:list,
%!         -MaxConOut:list, +MinIncIn:list, -MinIncOut:out,
%!         Direction:up_or_down_or_both) is det
%
% Explore potential blends for consistency. A blend is given by the base
% theory Base and some additional axioms from Axioms.
%
% The exploration starts with a given blend (Theory) and the goal is
% to find maximal consistent theories that include Theory
% and minimal inconsistent theories that are subsets of Theory.
% Theory can be expanded by formulas from Axioms (if those are not
% already part of Theory), or shrinked by removing formulas from Axioms
% (if they are part of Theory).
%
% The exploration can be supported by providing lists of blends that
% are already known to be consistent (MaxConIn) or inconsistent (MinIncIn).
%
% @arg Base
%   A backgound theory, i.e. a list of formula that are assumed.
%
% @arg Axioms
%   A set of axioms that may be part of the blend. Every
%   blend will be a subset of Axioms.
%
% @arg Theory
%   A blend to start exploration.
%
% @arg MaxConIn
%   A set of blends known to be consistent before this exploration step.
%
% @arg MaxConOut
%   A set of blends known to be consistent after this exploration step.
%
% @arg MinIncIn
%   A set of blends known to be inconsistent before this exploration step.
%
% @arg MinIncOut
%   A set of blends known to be inconsistent after this exploration step.
%
% @arg Direction
%   Direction of exploration (up, down, or both)


% T is contained in a consistent set, hence consistent.
explore(B, Ax, T, MaxConIn, MaxConOut, MinIncIn, MinIncOut, Direction) :-
    subset_of_sets(T, MaxConIn), !,
%    write('covered by MaxCon: '), write(MaxConIn), nl,
    explore_up(B, Ax, Ax, T, MaxConIn, MaxConOut, MinIncIn, MinIncOut, Direction).

% T is containing an inconsistent set, hence inconsistent.
explore(B, Ax, T, MaxConIn, MaxConOut, MinIncIn, MinIncOut, Direction) :-
    superset_of_sets(T, MinIncIn), !,
%    write('covered by MinInc: '), write(MinIncIn), nl,
    explore_down(B, Ax, [], T, MaxConIn, MaxConOut, MinIncIn, MinIncOut, Direction).

explore(B, Ax, T, _, _, _, _, Direction) :-
    write(exploring(B,Ax,T, Direction)), write(': '), flush_output, fail.

% Otherwise, we have to check if T is consistent.
explore(B, Ax, T, MaxConIn, MaxConOut, MinIncIn, MinIncOut, Direction) :-
    append(B, T, Theory),
    write('checking_consistency: '), flush_output,
    consistent(Theory), !,
    write('consistent!'), nl,
    superset_extend(MaxConIn, T, MaxConHelp),
    explore_up(B, Ax, Ax, T, MaxConHelp, MaxConOut, MinIncIn, MinIncOut, Direction).

explore(B, Ax, T, MaxConIn, MaxConOut, MinIncIn, MinIncOut, Direction) :-
    write('inconsistent!'), nl,
    subset_extend(MinIncIn, T, MinIncHelp),
    explore_down(B, Ax, [], T, MaxConIn, MaxConOut, MinIncHelp, MinIncOut, Direction).




%! explore_up(+Base:list, +Axioms:list, +Try:list, +Theory:list,
%!            +MaxConIn:list_of_lists, -MaxConOut:list_of_lists,
%!            +MinIncIn:list_of_lists, -MinIncOut:list_of_lists,
%!            +Direction:up_or_down) is det
%
% upward exploration: extend the (consistent) theory T by another axiom
% from Ax and check for consistency.
%
% @arg Base
%   A backgound theory, i.e. a list of formula that are assumed.
%
% @arg Axioms
%   A set of axioms that may be part of the blended theory. Every
%   blend theory T will be a subset of Axioms.
%
% @arg Try
%   A theory to explore.
%
% @arg Theory
%   The (consistent) theory to be explored.
%
% @arg MaxConIn
%   A set of blends known to be consistent before this exploration step.
%
% @arg MaxConOut
%   A set of blends known to be consistent after this exploration step.
%
% @arg MinIncIn
%   A set of blends known to be inconsistent before this exploration step.
%
% @arg MinIncOut
%   A set of blends known to be inconsistent after this exploration step.
%
% @arg Direction
%   Direction of exploration (up, down, or both)
explore_up(_, _, _, _, MaxCon, MaxCon, MinInc, MinInc, down) :-
    !.

% no more axioms left
explore_up(_, _, [], _, MaxCon, MaxCon, MinInc, MinInc, _) :-
    !.

% if axiom A is already in theory T, skip it an continue with the rest
explore_up(B, Ax, [A|R], T, MaxConIn, MaxConOut, MinIncIn, MinIncOut, Direction) :-
    element_of_set(A,T), !,
    explore_up(B, Ax, R, T, MaxConIn, MaxConOut, MinIncIn, MinIncOut, Direction).

explore_up(B, Ax, [A|R], T, MaxConIn, MaxConOut, MinIncIn, MinIncOut, Direction) :-
    explore(B, Ax, [A|T], MaxConIn, MaxConHelp, MinIncIn, MinIncHelp, up),
    explore_up(B, Ax, R, T, MaxConHelp, MaxConOut, MinIncHelp, MinIncOut, Direction).


%! explore_down(+Base:list, +Ax:list, +T1:list, +T2:list,
%!              +MaxConIn:list_of_lists, -MaxConOut:list_of_lists,
%!              +MinIncIn:list_of_lists, -MinIncOut:list_of_lists,
%!              +Direction:up_or_down) is det
%
% Downward exploration: remove a formula from the (inconsistent) theory
% {T1,T2} and check for consistency of the remaining formulae.
%
explore_down(_, _, _, _, MaxCon, MaxCon, MinInc, MinInc, up) :- !.

% no more axioms left.
explore_down(_, _, _, [], MaxCon, MaxCon, MinInc, MinInc, _) :- !.

explore_down(B, Ax, T1, [A|T2], MaxConIn, MaxConOut, MinIncIn, MinIncOut, Direction) :-
    append(T1,T2,T),
    explore(B, Ax, T, MaxConIn, MaxConHelp, MinIncIn, MinIncHelp, down),
    explore_down(B, Ax, [A|T1], T2, MaxConHelp, MaxConOut, MinIncHelp, MinIncOut, Direction).

