:- module(prover9, [prover9_consistent/1, prover9_consistent/2]).


/** <module> Prover9 interface
 *
 *    This module provides a predicate prover9_consistent/1 that
 *    can check if a theory, i.e. a set of first-order logical formulae
 *    is consistent.
 *
 *    The library uses "prover9" to accomplish this task. For more
 *    information on "prover9", especially for the expected syntax
 *    of formulae, see [[Prover9][http://www.cs.unm.edu/~mccune/mace4/]]
 *    In order to use this module, prover9 has to be downloaded and installed
 *    from the above URL. You may have to adapt the configuration data below
 *    to fit your installation.
 *
 *  @author Ulf Krumnack
 *  @copyright 2014
 *  @license GPL
 */


% Configuration
% -------------

% binary: name of the prover9 binary. If not located in the PATH, the
% full path to the binary has to be given here.
prover9_config('binary', 'prover9').

% timeout: Timeout in seconds. 
prover9_config('timeout', '2').

% verbosity: verbosity of operation.
%  0: no output
%  1: output status
%  2: output input theory
%  3: output proof(s)
prover9_config('verbosity', 0).

% communication: how to communicate with the theorem prover.
%  pipe: write to and read from pipe
%  *: write input and output to files, the given value is used as a basename
%     for the files, to which ".in", ".out", and ".err" is appended
prover9_config('communication', pipe).




string_list_concat("", [], _) :- !.

string_list_concat(String, [String], _) :- !.

string_list_concat(String, [First|Rest], Separator) :-
    string_concat(First, Separator, String1),
    string_list_concat(String2,Rest,Separator),
    string_concat(String1, String2, String).


%! write_theory(+Stream,+Theory:list) is det
%
% Write the given theory to a stream that can be processed by prover9.
%
% @param Stream a Prolog stream.
% @param Theory a list of formulas in a format understood by prover9.
write_theory(Stream, Theory) :-
    prover9_config('timeout', Timeout),
    string_list_concat(TimeoutStr, ['assign(max_seconds, ', Timeout, ').'], ''),
    write(Stream, TimeoutStr), nl(Stream),
    write(Stream, 'formulas(sos).'), nl(Stream),
    write_theory_help(Stream, Theory),
    write(Stream, 'end_of_list.'), nl(Stream).

write_theory_help(_, []) :- !.

write_theory_help(Stream, [First|Rest]) :-
    write(Stream, '  '),
    write(Stream, First),
    nl(Stream),
    write_theory_help(Stream, Rest).



%! filter_output(+Output:stream) is det
%
% Analyze the output of the prover9 process.
%
% This predicate is a stub. It may be adapted in future to do something
% sensible.
filter_output(Output) :-
    % read_string(Output, "\n", "\r", _, Line),
    read_line_to_codes(Output, Codes),
    filter_output_line(Output, Codes).

filter_output_line(_, end_of_file) :- !.

filter_output_line(Output, Codes) :-
    prover9_config('verbosity', Verbosity), Verbosity >= 3,
    atom_codes(Line, Codes),
    sub_string(Line, _, _, _, 'PROOF'), !,
    write(user_error, Line), nl(user_error),
    filter_output(Output).

filter_output_line(Output, _) :-
    filter_output(Output).



%! prover9_consistent(+Theory:list) is semidet
%
% Check if the given Theory is consistent.
%
% @arg Theory
%   is expected to be a list of formulae, given as strings in a
%   format understood by prover9.
prover9_consistent(Theory) :-
    prover9_consistent(Theory,Status),
    Status \== 0.

%! prover9_consistent(+Theory:list, -Status:int) is det
%
% Check if the given Theory is consistent.
%
% @arg Theory
%   is expected to be a list of formulae, given as strings in a
%   format understood by prover9.
%
% @arg Status
%   is the exit status of the prover9 process.
%   It will be 0 (MAX_PROOFS), if the maximal  number of proofs
%   have been found. This means, that the inconsistency of Theory
%   was proven. All other exit codes mean, that the inconsistency
%   of Theory was not proven.
%   1 (FATAL) indicates an error. This is typically caused by a syntax
%     error in Theory.
%   2 (SOS_EMPTY) means that the list of tasks is empty,
%     which indicates, that Theory is consistent.
%   3 (MAX_MEGS) and 4 (MAX_SECONDS) means that some resource limit
%     was reached without finding a contradiction, which makes it
%     probable that the theory is consistent.
%   Other exit codes indicate that the search was interrupted before an
%   inconsistency was found.
prover9_consistent(Theory,Status) :-
    prover9_config('verbosity', Verbosity),
    ( Verbosity >= 2 -> write_theory(user_error, Theory) ; true ),
    prover9_config('communication', Communication),
    ( Communication == pipe ->
      prover9_consistent_pipe(Theory, Status) ;
      prover9_consistent_file(Theory, Status, Communication)),
    ( Status == 1 -> throw('Syntax error!') ; true ),
    ( Verbosity >= 1 ->
      write(user_error, status = Status), nl(user_error) ; true ).


% Implementation 1 (Pipe)
% -----------------------
%
% Run prover9 as a background process, communication is performed using
% pipes. The following steps are perfomed:
% 1. Run prover9 as a background process, using pipes to send input
%    data and read output data (using process_create/3).
% 2. Send the theory to check for consistency to the input stream of
%    the prover9 process
% 3. Read and evaluate the output of the prover9 process (optional).
% 4. Determine the exit code of the prover9 process.
    
prover9_consistent_pipe(Theory, Status) :-
    prover9_config('binary', Prover9),
    process_create(path(Prover9), [ ],
                   [ stdin(pipe(Input)), stdout(pipe(Output)),
                     stderr(null), process(PID)]),
    write_theory(Input, Theory),
    close(Input),
    filter_output(Output),
    is_process(PID),
    process_wait(PID, exit(Status)), !.

prover9_consistent_pipe(_, _) :-
    throw('Error running theorem prover!').


% Implementation 2 (Files)
% ------------------------
%
% Run prover9 as a shell command, reading input from and writing output
% to temporary files. The following steps are perfomed:
% 1. prepare: the theory is written into a file (suffix: ".in")
% 2. perform: prover9 is run on that file, redirecting its output into
%    two other files (suffix ".out" and ".err") using shell/2
% 3. evaluate (optional): the output is analyzed, e.g. to provide a proof


prover9_consistent_file(Theory, Status, Filename) :-
    prover9_config('binary', Prover9),
    string_concat(Filename, '.in', InFile),
    string_concat(Filename, '.out', OutFile),
    string_concat(Filename, '.err', ErrFile),
    open(InFile, write, InStream),
    write_theory(InStream, Theory),
    close(InStream),
    string_list_concat(Command,
                       [Prover9, '<', InFile, '>', OutFile, '2>', ErrFile],
                       ' '),
    shell(Command, Status),
    open(OutFile, read, OutStream),
    filter_output(OutStream),
    close(OutStream),
    delete_file(InFile),
    delete_file(OutFile),
    delete_file(ErrFile), !.

prover9_consistent_file(_, _, _) :-
    throw('Error running theorem prover!').
